/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#ifndef SQLUTILS_H
#define SQLUTILS_H

#include <iostream>
#include <cmath>
#include <cstring>
#include <string>
#include <vector>
#include <map>
#include <arpa/inet.h>
#include <cstdlib>
using namespace std;

class restuple {
public:
	vector <string> precolvals;	//the list of attr vals in the result based on input query
	double matchwt;		//the -log prob of the output, ie -log(sum of probs of matched lines)
	restuple() {
		matchwt = 0.0;
	}
	~restuple(){}
};

enum SQLType {KMAP, FULLFST, STAC, STACPROJ};

inline void deathmsg(string msg) {
	cout << msg;
	exit(1);
}

class arguments {
public:
	string dbname, portnum, sqlquery, regex, indexterm;
	int K, M, L, V;
	arguments(){
		K = M = L = -1;
		V = 0;
		dbname = portnum = sqlquery = regex = indexterm = string("");
	};
	~arguments(){};
	void populate (int argc, char **argv) {
		dbname = string(argv[1]);
		portnum = string(argv[2]);
		sqlquery = string(argv[3]);
		regex = string(argv[4]);
		int argi = 5;
		while(argi < argc) {
			if(strcmp(argv[argi], "-k") == 0)
				K = atoi(argv[argi + 1]);
			if(strcmp(argv[argi], "-m") == 0)
				M = atoi(argv[argi + 1]);
			if(strcmp(argv[argi], "-i") == 0)
				indexterm = string(argv[argi + 1]);
			if(strcmp(argv[argi], "-v") == 0)
				V = 1;
			if(strcmp(argv[argi], "-l") == 0)
				L = atoi(argv[argi + 1]);
			argi++;
		}
	}
	void check (SQLType sqltype) {
		switch(sqltype) {
			case KMAP:
				if(K == -1) deathmsg("-k <k value> not provided\n");
				break;
			case STACPROJ:
				if(indexterm == "")	deathmsg("-i <index term> not provided\n");
			case STAC:
				if(K == -1)	deathmsg("-k <k value> not provided\n");
				if(M == -1) deathmsg("-m <m value> not provided\n");
				break;
		}
	}
};

class querystructs {
public:
	//iterate thro the query toks to get the structs we need
	vector <string> auxcolnames;
	string fstcolname;
	string tablename;
	string regex;
	int Ktop;
	int type;	//0 for answer/extraction, 1 for exists type query
	querystructs(){};
	~querystructs(){};
	string constructCmd() {
		string cmdstring = "select " + fstcolname + ", " + fstcolname + "id";
		for(int c=0; c < auxcolnames.size(); c++) {
			cmdstring += ", " + auxcolnames[c];
		}
		cmdstring += " from " + tablename + ";";	//first we get the keys; then we iterate thro them in the for loop; the master table also has the aux cols
		return cmdstring;
	}
	string constructCmdId(string thisid) {
		string cmdstring = constructCmd();
		cmdstring = cmdstring.substr(0, cmdstring.length()-1); //remove ';'
		cmdstring += " where " + fstcolname + "id = " + thisid + ";";
		return cmdstring;
	}
};

void printUsage (SQLType sqltype);
int unpack_int32(char *data);
double unpack_double(char* data);
int dfacheck (string inpstr, int* dfaquery, int inpstate, int NUMST, int startstate, int finalstate);
void sortResults (vector <restuple> *results);
void printResults (vector <restuple> results, vector <string> auxcolnames, int numAns);
querystructs parseQuery(char* sqlquery);

#endif
