/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
        
This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/                      
#include "approximation/approxfastUtils.h"

int VERBOSE = -1;

void fixMinMaxSkel(blockskel *thisskel, map <int, set<int> > *adjlist, int* nodedismin, int numnodes, int minmax) {
	//if any of the skel nodes have outvxs set of size 0, it is _a_ maxnode
	//the out vals would corresp to in vals, if adjlist corresp to reversed edges in orig fst
	if(VERBOSE >= 2) {
	cout << "\t\tin fixminmax with minmax " << minmax << ", skel.sfa:\n";
	for(map <int, set <int> >::iterator iter = (thisskel->sfa).begin(); iter != (thisskel->sfa).end(); iter++) {
		cout << iter->first << ":";
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++) {
			cout << " " << *siter;
		}
		cout << endl;
	}
	}
			
	set <int> extremenodes;
	int maxextrnode = -1, minextrnode = numnodes;
	if(minmax == 1) { //need to check maxnode, i.e. outdeg
		for(map <int, set <int> >::iterator iter = (thisskel->sfa).begin(); iter != (thisskel->sfa).end(); iter++)
			if(iter->second.size() == 0) {
				extremenodes.insert(iter->first);
				if(VERBOSE >= 2) cout << "\tfor max, adding " << iter->first << " as extrnode\n";
				maxextrnode = (maxextrnode < iter->first) ? iter->first : maxextrnode;
				minextrnode = (minextrnode > iter->first) ? iter->first : minextrnode;
			}
	}
	else { //need to check minnode, i.e. indeg
		for(map <int, set <int> >::iterator iter = (thisskel->sfarev).begin();  iter != (thisskel->sfarev).end(); iter++)
			if(iter->second.size() == 0) {
				extremenodes.insert(iter->first);
				if(VERBOSE >= 2) cout << "\tfor min, adding " << iter->first << " as extrnode\n";
				maxextrnode = (maxextrnode < iter->first) ? iter->first : maxextrnode;
				minextrnode = (minextrnode > iter->first) ? iter->first : minextrnode;
			}
	}
	
	if(VERBOSE >= 2) {
	cout << "\t\tin fixminmax, extrnodes:";
	for(set <int>::iterator siter = extremenodes.begin(); siter != extremenodes.end(); siter++)
		cout << " " << *siter;
	cout << endl;
	}
	
	if(extremenodes.size() == 1) {
		if(minmax == 1)
			thisskel->maxnode = *(extremenodes.begin());
		else
			thisskel->minnode = *(extremenodes.begin());
	}
	else if(extremenodes.size() < 1) {
		cout << "Error: fixMinMaxSkel: No min/max extreme nodes!\n";
		exit(1);
	}
	else {	//more than one extrnode; go thro list of nodes in full sfa, and see which ones are reachable from all extrnodes, then get opp. extr among those
		//we only need to start counting in the full sfa from the max/min extrnode! go fwds or bkwds dep on min/max; then we stop at the first one which is a meetpt!
		int meetpt;
		if(minmax == 1) { //fix max node
			for(int anode = maxextrnode; anode < numnodes; anode++) {	//the meetpt could be an extrnode itself!
				if(adjlist->find(anode) == adjlist->end())	//this node is not present in the last sane full sfa
					continue;
				int rchblflag = 1; //alter the rchblty info in the skel for loop, by anding for all nodes in endset
				for(set <int>::iterator siter = extremenodes.begin(); siter != extremenodes.end(); siter++) {
					if(nodedismin[(*siter) * numnodes + anode] == INFIN) {
						rchblflag = 0;
						break;	//anode is not reachble from at least one extrpt, so ignore!
					}
				}
				if(rchblflag == 1) {
					meetpt = anode;
					thisskel->maxnode = meetpt;
					break;	//this is the smallest anode that is rchbl from all extrpts, and since sfa is toposorted, no need to look further!
				}
			}
		}
		else { //fix min mode
			for(int anode = minextrnode; anode >= 0; anode--) {		//the meetpt could be an extrnode itself!
				if(adjlist->find(anode) == adjlist->end())	//this node is not present in the last sane full sfa
					continue;
				int rchblflag = 1; //alter the rchblty info in the skel for loop, by anding for all nodes in endset
				for(set <int>::iterator siter = extremenodes.begin(); siter != extremenodes.end(); siter++) {
					if(nodedismin[anode * numnodes + (*siter)] == INFIN) {
						rchblflag = 0;
						break;	//anode is not reachble from at least one extrpt, so ignore!
					}
				}
				if(rchblflag == 1) {
					meetpt = anode;
					thisskel->minnode = meetpt;
					break;	//this is the largest anode that is revrchbl from all extrpts, and since sfa is toposorted, no need to look further!
				}
			}
		}
		
		//now based on the meetpt, we can do the fix -> basically, add all nodes/edges that lie on a path from any of the extreme nodes and the meetpt
		//if we are fixing maxnode, we only need to look from minextrnode to meetpt on the fullgph! since it is toposorted, we wont hv other nodes in the expansion!
	
		if(VERBOSE >= 2) cout << "\t\tin fixminmax added meetpt " << meetpt << endl;
		(thisskel->sfa)[meetpt];				(thisskel->sfarev)[meetpt];  //add meetpt to sub sfa
		
		if(minmax == 1) { //fix max node
			for(int anode = minextrnode + 1; anode <= meetpt; anode++) {
				if(adjlist->find(anode) == adjlist->end())	//not present in the last sane full sfa
					continue;
				for(set <int>::iterator siter = extremenodes.begin(); siter != extremenodes.end(); siter++)
					if((nodedismin[(*siter) * numnodes + anode] != INFIN) && (nodedismin[anode * numnodes + meetpt] != INFIN)) {
						(thisskel->sfa)[anode];			 (thisskel->sfarev)[anode]; 
						if(VERBOSE >= 2) cout << "\t\t\tadded node to subsfa: " << anode << endl;
					}
			}
		}
		else { //fix min node
			for(int anode = maxextrnode - 1; anode >= 0; anode--) {
				if(adjlist->find(anode) == adjlist->end())	//not present in the last sane full sfa
					continue;
				for(set <int>::iterator siter = extremenodes.begin(); siter != extremenodes.end(); siter++)
					if((nodedismin[anode * numnodes + (*siter)] != INFIN) && (nodedismin[meetpt * numnodes + anode] != INFIN)) {
						(thisskel->sfa)[anode];			 (thisskel->sfarev)[anode]; 
						if(VERBOSE >= 2) cout << "\t\t\tadded node to subsfa: " << anode << endl;
					}
			}
		}		
		
		//get all edges among any pair of nodes in the newly augmented subsfa (it has assimiltd the newly inducted nodes)
		//while updating overall strucut later, we drop all edges of this sub sfa and drop all non-extreme nodes from overall strucuts, and do collapse addn 
		for(map <int, set <int> >::iterator iter = (thisskel->sfa).begin(); iter != (thisskel->sfa).end(); iter++) {
			for(map <int, set <int> >::iterator siter = (thisskel->sfa).begin(); siter != (thisskel->sfa).end(); siter++) {
				if(iter->first == siter->first)
					continue;
				if((*adjlist)[iter->first].find(siter->first) != (*adjlist)[iter->first].end()) {	//iter to siter edge exists, we do n^2-n, not nc2
					(thisskel->sfa)[iter->first].insert(siter->first);				(thisskel->sfarev)[siter->first].insert(iter->first);
					if(VERBOSE >= 2) cout << "\t\t\tadded edge to subsfa: " << iter->first << "-" << siter->first << endl;
				}
			}
		}				
				
	}//end else, for fix opertn
}


void fixMidlSkel(blockskel *thisskel, map <int, set <int> > *adjlist, map <int, set <int> > *adjlistrev) {
	if(VERBOSE >= 2) {
	cout << "\t\tin fixmidlskel, skel.sfa has min,max: " << thisskel->minnode << "," << thisskel->maxnode << " and sfa:\n";
	for(map <int, set <int> >::iterator iter = (thisskel->sfa).begin(); iter != (thisskel->sfa).end(); iter++) {
		cout << iter->first << ":";
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++) {
			cout << " " << *siter;
		}
		cout << endl;
	}
	}
	
	set < Edge > accumulated;
	//go over the non-extreme nodes in the sub sfa, and check if any such node has an edge in adjlist with the other vertex outside of the sub sfa, or reverse too!
	for(map <int, set <int> >::iterator iter = (thisskel->sfa).begin(); iter != (thisskel->sfa).end(); iter++) {
		if((iter->first == thisskel->maxnode) || (iter->first == thisskel->minnode))
			continue;
		for(set <int>::iterator siter = (*adjlist)[iter->first].begin(); siter != (*adjlist)[iter->first].end(); siter++) {
			if((thisskel->sfa).find(*siter) == (thisskel->sfa).end()) {	//so the other vx is not inside sub sfa, this edge needs to be assimilated later
				//(thisskel->sfa)[iter->first].insert(*siter);				(thisskel->sfarev)[*siter].insert(iter->first);	//BUG! cannot modify sfa inside this loop
				//(thisskel->sfa)[*siter];					(thisskel->sfarev)[iter->first];
				accumulated.insert(make_pair(iter->first,*siter));
				if(VERBOSE >= 2) cout << "\t\t\tadded node to subsfa: " << *siter << " due to edge to subsfa: " << iter->first << "-" << *siter << endl;
			}
		}
		for(set <int>::iterator siter = (*adjlistrev)[iter->first].begin(); siter != (*adjlistrev)[iter->first].end(); siter++) {
			if((thisskel->sfa).find(*siter) == (thisskel->sfa).end()) {	//so the other vx is not inside sub sfa, this edge needs to be assimilated later
				//(thisskel->sfa)[iter->first].insert(*siter);				(thisskel->sfarev)[*siter].insert(iter->first);	//BUG! cannot modify sfa inside this loop
				//(thisskel->sfa)[*siter];					(thisskel->sfarev)[iter->first];
				accumulated.insert(make_pair(*siter,iter->first));
				if(VERBOSE >= 2) cout << "\t\t\tadded node to subsfa: " << *siter << " due to edge to subsfa: " << *siter << "-" << iter->first << endl;
			}
		}
	}
	//assimilate the accumulated edges
	for(set < Edge >::iterator eiter = accumulated.begin(); eiter != accumulated.end(); eiter++) {
		(thisskel->sfa)[eiter->first].insert(eiter->second);					(thisskel->sfarev)[eiter->second].insert(eiter->first);
		(thisskel->sfa)[eiter->second];						(thisskel->sfarev)[eiter->first];
	}
	//if no edges to accumulate, we finally need to check if there is an edge from min to max, if so, add it -> *** a key distcn in blk defn from before!
	if((*adjlist)[thisskel->minnode].find(thisskel->maxnode) != (*adjlist)[thisskel->minnode].end()) {
		(thisskel->sfa)[thisskel->minnode].insert(thisskel->maxnode);				(thisskel->sfarev)[thisskel->maxnode].insert(thisskel->minnode);
		if(VERBOSE >= 2) cout << "\t\t\tadded key distcn edge to subsfa: " << thisskel->minnode << "-" << thisskel->maxnode << endl;
	}
}

blockskel findMinSfa(map <int, set <int> > *adjlist, map <int, set <int> > *adjlistrev, int* nodedismin, int numnodes, pair<Edge, Edge> thepair) {
	blockskel mergedskel;
	mergedskel.sfa[thepair.first.first].insert(thepair.first.second);	mergedskel.sfa[thepair.second.first].insert(thepair.second.second);
	mergedskel.sfa[thepair.first.second]; 	mergedskel.sfa[thepair.second.second];
	mergedskel.sfarev[thepair.first.second].insert(thepair.first.first); mergedskel.sfarev[thepair.second.second].insert(thepair.second.first);
	mergedskel.sfarev[thepair.first.first];	mergedskel.sfarev[thepair.second.first];
	while(1) {	//keep iterating and fixing as long as the cand blk is illegal
		blockskel copyskel = mergedskel;	//to be used at end of itern to check if there has been any change
		fixMinMaxSkel(&mergedskel, adjlist, nodedismin, numnodes, 1);		//fix maxnode if needed
		fixMinMaxSkel(&mergedskel, adjlist, nodedismin, numnodes, -1);		//fix minnode if needed
		fixMidlSkel(&mergedskel, adjlist, adjlistrev);
		if (mergedskel == copyskel) //no more cahnges - its now a legal candidate
			break;
	}
	return mergedskel;
}	

bool pathwtcomp (pair<string,double> fir, pair<string,double> sec) {
	return (fir.second < sec.second) ? true : false;
}


double probDynMult(map <int, set <int> > *adjlist, map < Edge, block> *blockdata, int startnode, map <int, double> *memo) {
	if(memo->find(startnode) != memo->end())
		return (*memo)[startnode];
	if((*adjlist)[startnode].size() == 0) //no children
		return ((*memo)[startnode] = 1.0);
	else { //obtain sum of children's prob funcs, multiplied with the blkprob on that edge
		double sumchildren = 0.0;
		for(set<int>::iterator citer = (*adjlist)[startnode].begin(); citer != (*adjlist)[startnode].end(); citer++)
			sumchildren += (*blockdata)[make_pair(startnode, *citer)].totprob * probDynMult(adjlist, blockdata, *citer, memo);
		return ((*memo)[startnode] = sumchildren);
	}
}


//get the top K paths and probs for the given subfst block -> a legal block; so we can iterate the origfstnodes set to build statepr; scratch is assumed to be clear!
block getBlockData(blockskel thisskel, map <Edge, block> *blockdata, int K, int storepaths, int* scratch) {
	block retblk;
	PathIterator incrementalN;
	constructIteratorNoFst(&incrementalN, thisskel, blockdata, scratch);
	if(VERBOSE >= 2) cout << "\tconstructed iterator (nofst)\n";
	for(int j=0;j < K; j++)	{
		Path p = incrementalN.getNext();
		if(incrementalN.moreflag == 0)	//no more paths
			break;
		double cost = p.getPathCost();
		retblk.totprob += exp(-1.0*cost);	//prob is e^-wt
		if(storepaths == 1)
			retblk.pathwts.push_back(make_pair(p.getPathStr(),cost));
	}
	delete incrementalN.enumerator;
	if(VERBOSE >= 2) cout << "\tretblk totprob " << retblk.totprob << endl;
	return retblk;
}


//use thisskel to do the collapse opern - throw out the interior nodes and all edges of the skel from the overall structs,  add in a single edge for the skel
void updateStructures(blockskel thisskel, block *thisblk, map <int, set<int> > *tmpadjlist, map <int, set<int> > *tmpadjlistrev, map <Edge, block> *tmpblockdata) {
	for(map <int, set <int> >::iterator iter = (thisskel.sfa).begin(); iter != (thisskel.sfa).end(); iter++) {
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++) {	//drop all the edges of the skel from the overall
			(*tmpblockdata).erase(make_pair(iter->first, *siter));
			(*tmpadjlist)[iter->first].erase(*siter);							(*tmpadjlistrev)[*siter].erase(iter->first);
		}
		if((iter->first != thisskel.minnode) && (iter->first != thisskel.maxnode)) {	//drop all interior nodes of the skel from the overall
			(*tmpadjlist).erase(iter->first);								(*tmpadjlistrev).erase(iter->first);
		}
	}
	(*tmpblockdata)[make_pair(thisskel.minnode,thisskel.maxnode)] = *thisblk;	//add min to max edge to complete the collapse operation
	(*tmpadjlist)[thisskel.minnode].insert(thisskel.maxnode);				(*tmpadjlistrev)[thisskel.maxnode].insert(thisskel.minnode);			//add edge to adj lists too
}


set < Edge > getEdgeSet(map <int, set <int> > sfa) {
	set < Edge > ret;
	for(map <int, set <int> >::iterator iter = sfa.begin(); iter != sfa.end(); iter++)
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++)
			ret.insert(make_pair(iter->first, *siter));
	return ret;
}

set < Edge > intersectEdgeSets(set < Edge > s1, set < Edge > s2) {
	set < Edge > ret;
	std::set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(ret, ret.begin()));
	return ret;
}
