/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "frontends/staccatoSQLIndexUtils.h"

//compute sumprob of parents by doing dyn prog rev dfs; thus, in case 2, intersblkprob can be simply added to [0][0] of parsumprob before mult at intersblks
void dynmatrdfs (map <int, vector <int> > *revgph, vector<double>* blkreachprobs, double* probmatrs, double* scratches, int* memo, int NUMST, int curnode, map<int,int>* newblkno) {
	int ncurnode = (*newblkno)[curnode];
	if (memo[ncurnode] == 1)
		return;
	int matrsize = NUMST * NUMST;
	if((*revgph)[curnode].size() == 0) { //if no pars, it is a global startblk! so mult its blkreachprob in the probmatr and memoize this, and return
		scalarmultmatr(probmatrs + (ncurnode * matrsize), NUMST, (*blkreachprobs)[curnode]);
		memo[ncurnode] = 1;
		return;
	}
	for(vector <int>::iterator iter = (*revgph)[curnode].begin(); iter != (*revgph)[curnode].end(); iter++) { //dfs; visit all parents
		if (memo[(*newblkno)[*iter]] == 1)
			continue;
		dynmatrdfs (revgph, blkreachprobs, probmatrs, scratches, memo, NUMST, *iter, newblkno);
	}
	memset(scratches, 0, 2 * matrsize * sizeof(double));	//clear the scratches; size of each prob matr is NUMST * NUMST * sizeofdouble in bytes
	for(vector <int>::iterator iter = (*revgph)[curnode].begin(); iter != (*revgph)[curnode].end(); iter++) {
		int paren = (*newblkno)[*iter];
		incrsummatrs(scratches, probmatrs + (paren * matrsize), NUMST);	//add this par's probmatr to the scratch0
	}
	//write to scratch1, the scratch0 mult with prev matr of curnodes! then, rewrite with memcpy to curnode's matr, memoize this and return
	multmatrs(scratches + matrsize, scratches, probmatrs + (ncurnode * matrsize), NUMST);	//mult is not commutativ; mult incrsumpar with curnode matr
	memcpy(probmatrs + (ncurnode * matrsize), scratches + matrsize, matrsize * sizeof(double));
	memo[ncurnode] = 1;
	return;
}

//similar as above, but add the disjreachprob to the inter MS startblks that occur while we process the nodes; add the disjreachprob to the [0][0] entry
void dynmatrdfsinter (map <int, vector <int> > *revgph, vector<double>* blkreachprobs, double* probmatrs, double* scratches, int* memo, int NUMST, int curnode, map<int,int>* newblkno, map<int, double> *intersblkprob) {
	int ncurnode = (*newblkno)[curnode];
	if (memo[ncurnode] == 1)
		return;
	int matrsize = NUMST * NUMST;
	if((*revgph)[curnode].size() == 0) {
		scalarmultmatr(probmatrs + (ncurnode * matrsize), NUMST, (*blkreachprobs)[curnode]);
		memo[ncurnode] = 1;
		return;
	}
	for(vector <int>::iterator iter = (*revgph)[curnode].begin(); iter != (*revgph)[curnode].end(); iter++) {
		if (memo[(*newblkno)[*iter]] == 1)
			continue;
		dynmatrdfs (revgph, blkreachprobs, probmatrs, scratches, memo, NUMST, *iter, newblkno);
	}
	memset(scratches, 0, 2 * matrsize * sizeof(double));
	for(vector <int>::iterator iter = (*revgph)[curnode].begin(); iter != (*revgph)[curnode].end(); iter++) {
		int paren = (*newblkno)[*iter];
		incrsummatrs(scratches, probmatrs + (paren * matrsize), NUMST);
	}
	//we have matr sum of parents; need to boost the [0][0] prob with the intersblkprob; this is done iff curnode is present in intersblkprob
	map<int, double>::iterator tmpsb = (*intersblkprob).find(curnode);
	if(tmpsb != (*intersblkprob).end())	//so this is an intersblk
		scratches[0] += tmpsb->second;

	multmatrs(scratches + matrsize, scratches, probmatrs + (ncurnode * matrsize), NUMST);
	memcpy(probmatrs + (ncurnode * matrsize), scratches + matrsize, matrsize * sizeof(double));
	memo[ncurnode] = 1;
	return;
}

//given this blknum and remlen, we construct the conservative estimate subgph and the endblks by doing bfs and using toplabs
void getprojn(structureInfo *stacinfo, map <int, vector <int> > *stacgraphrev, map<int, vector<int> > *stacgraphproj, set <int> *endblks, int remlen, int curblk) {
	(*stacgraphproj)[curblk];
	int numblks = (stacinfo->stacgraph).size();
	int rems[numblks];		//if it is -1, it means it is not to be used! if not, it gives the actual remlen for this blk
	bool present[numblks];		//this will indicate whetehr the blk is present in the visit queue or not
	for(int i=0; i<numblks; i++) {
		rems[i] = -1;
		present[i] = 0;
	}
	int currem = remlen;
	rems[curblk] = remlen;	//so far, the onyl legal blk
	present[curblk] = 1;
	priority_queue<int> visitq;	//maintain a maxheap on the -1*toplabel of the blknum; heap contains -toplabs; we need toplabsrev to get blknum back
	vector<int> children = (stacinfo->stacgraph)[curblk];
	int numchildren = children.size();
	//since we use priorq on topo labels, no need to sort children etc simply push children onto q
	for(vector<int>::iterator citer = children.begin(); citer != children.end(); citer++) {
		(*stacgraphproj)[*citer];
		(*stacgraphproj)[curblk].push_back(*citer);
		visitq.push(-1 * (stacinfo->toplabs)[*citer]);
		present[*citer] = 1;
	}
	//iterative bfs
	while(!visitq.empty()) {
		int thisblk = (stacinfo->toplabsrev)[-1 * visitq.top()];	//pop the min toplab and mult by -1
		visitq.pop();
		present[thisblk] = 1;
		//now, check this blk's parents and get their rems; then, find the max among those, and subtract this blk's minlen to get next remlen to pass on
		int maxparrem = 0;
		for(vector<int>::iterator parit = (*stacgraphrev)[thisblk].begin(); parit != (*stacgraphrev)[thisblk].end(); parit++) {
			if(present[*parit] != 1) //this parent is not present in the subgph we are building, ignore
				continue;
			if(maxparrem < rems[*parit])
				maxparrem = rems[*parit];
		}
		int passonrem = maxparrem - (stacinfo->mmlens)[thisblk].first;		//max par remlen - this blk's minlen
		if(passonrem <= 0) {	//no need to visit more children; also add to end nodes
			(*endblks).insert(thisblk);
			rems[thisblk] = 0;
		}
		else { //add passonrem to rems and then add children to visit queue; also add links to proj gph; note that children have to be added in topo sort order
			rems[thisblk] = passonrem;
			vector<int> tchildren = (stacinfo->stacgraph)[thisblk]; //again, no need to sort children since this is priorq
			int tnumchildren = tchildren.size();
			for(vector<int>::iterator tciter = tchildren.begin(); tciter != tchildren.end(); tciter++) {
				(*stacgraphproj)[*tciter];
				(*stacgraphproj)[thisblk].push_back(*tciter);
				if(present[*tciter] == 1)		//shouldnt add to priorq if already exists; but the projgph shld incl this link
					continue;
				visitq.push(-1 * (stacinfo->toplabs)[*tciter]);
				present[*tciter] = 1;
			}
		}//end if on passonrem
	}//end of bfs
}

//var len query; but single start blk; so get subph till end by dfs
void dfstillend(map<int, vector<int> > *stacgraph, map<int, vector<int> > *stacgraphproj, int* allcdone, int curblk) {
	if((*stacgraph)[curblk].size() == 0) {		//no children, so add empty list
		vector<int> emptyv;
		(*stacgraphproj)[curblk] = emptyv;
		allcdone[curblk] = 1;
		return;
	}
	for(vector<int>::iterator iter = (*stacgraph)[curblk].begin(); iter != (*stacgraph)[curblk].end(); iter++) {
		(*stacgraphproj)[curblk].push_back(*iter); //add this node to child link!
		if(allcdone[*iter] == 0)
			dfstillend(stacgraph, stacgraphproj, allcdone, *iter);
	}
	allcdone[curblk] = 1;
}

void dfsgetreach(map<int, vector<int> > *gph, int curblk, int* reachable) {	//reachable is assumed to be an array of length numblks!
	//already marked reachable since a dfs visit was completed before; this enables sharing across starting from each MStop node when computing afteMS; or reverse too
	if(reachable[curblk] == 1)
		return;
	for(vector<int>::iterator iter = (*gph)[curblk].begin(); iter != (*gph)[curblk].end(); iter++) {
		if(reachable[*iter] == 1)		//already reached this child before; so, no need to visit again
			continue;
		dfsgetreach(gph, *iter, reachable);
	}
	reachable[curblk] = 1;
}

void dfsgetSBi(structureInfo *stacinfo, int *curtops, int *memor, set <int> *SBi, int curblk, int limitblk, int sign) {
	if(curtops[(stacinfo->toplabs)[curblk]] == 1)		//this blk is present in the curtops set
		(*SBi).insert((stacinfo->toplabs)[curblk]);
	if((sign == 1) && ((stacinfo->toplabs)[curblk] >= (stacinfo->toplabs)[limitblk])) {	//this is the upper limit
		memor[curblk] = 1;
		return;
	}
	else if((sign == -1) && ((stacinfo->toplabs)[curblk] <= (stacinfo->toplabs)[limitblk])) {	//this is the lower limit
		memor[curblk] = 1;
		return;
	}
	for(vector<int>::iterator iter = (stacinfo->stacgraph)[curblk].begin(); iter != (stacinfo->stacgraph)[curblk].end(); iter++) {
		if(memor[*iter] == 1)		//child already fin dfs visit
			continue;
		dfsgetSBi(stacinfo, curtops, memor, SBi, *iter, limitblk, sign);
	}
	memor[curblk] = 1;	//dfs visit is fin, so set memo flag
}

//this is the prob of reaching a block without going thro any other node in the MStops (basically, its ancestors)
double dyndisjreachprob(map<int, vector<int> > *revgph, structureInfo* stacinfo, int curblk, set<int> *MStops, int* memodynflag, double* memodynprob) {
	set<int>::iterator tmpt = (*MStops).find((stacinfo->toplabs)[curblk]);
	if(tmpt != (*MStops).end()) {	//we reached an MS blk
		memodynprob[curblk] = 0.0;
		memodynflag[curblk] = 1;
		return 0.0;
	}
	double tmp;
	if((*revgph)[curblk].size() == 0)	//no parents, since this is a global start blk; so, make it I
		tmp = 1.0;
	else {
		for(vector<int>::iterator piter = (*revgph)[curblk].begin(); piter != (*revgph)[curblk].end(); piter++) {
			if(memodynflag[*piter] == 1)		//par fin dfs before
				tmp += memodynprob[*piter];
			else
				tmp += dyndisjreachprob(revgph, stacinfo, *piter, MStops, memodynflag, memodynprob);		//sum of probs sent from parents
		}
	}
	tmp *= (stacinfo->blkprobs)[curblk];
	memodynprob[curblk] = tmp;
	memodynflag[curblk] = 1;
	return tmp;
}

//get minimal set of startblks/endblks - flag -1 for endblks and +1 for startblks
void getMinSet(structureInfo stacinfo, set<int> extblks, set<int> *MSexttops, int flag) {
	int numblks = stacinfo.stacgraph.size();
	int numextblks = extblks.size();
	int eblktops[numextblks];		//the topolabs of the endblks; we sort this
	int ecurtops[numblks];			//the eblkstops presently under considern (set S); indexed by toplab of blkno
	memset((char*)ecurtops, '\0', sizeof(int)*numblks);
	int esindex = 0;
	for(set<int>::iterator eiter = extblks.begin(); eiter != extblks.end(); eiter++) {
		eblktops[esindex] = stacinfo.toplabs[*eiter];
		esindex++;
		ecurtops[stacinfo.toplabs[*eiter]] = 1;
	}
	sortArray(eblktops, numextblks);		//now we hv the list of eblk toplabs in sorted order; let us go from highest to lowest and obtain reachability info
	for(int z=0; z<numextblks; z++) {
		if(ecurtops[eblktops[z]] == 0) //check if this sblktop is present in curtops
			continue;				//now we have an eblktop in the curtops, and it is an MSEtop by algm (since the list is inv sorted in topo order)
		MSexttops->insert(eblktops[z]);		//shld be the toplabs; not origblknos
		//get list of nodes reachable from this node; we only need to go till greatest sblktopo in sblktops! pass along set intersxn (SBi) to be computed
		set <int> extSBi;		//include Bi itself too
		int memor[numblks];			//the eblkstops presently under considern (set S)
		memset((char*)memor, '\0', sizeof(int)*numblks);
		dfsgetSBi(&stacinfo, ecurtops, memor, &extSBi, stacinfo.toplabsrev[eblktops[z]], stacinfo.toplabsrev[eblktops[numextblks-1]], flag);
		for(set<int>::iterator esbiter = extSBi.begin(); esbiter != extSBi.end(); esbiter++) //substract extSBi from S; go thro extSBi and unset their ecurtops flag
			ecurtops[*esbiter] = 0;
	}
}

void getInterProb(structureInfo stacinfo, map <int, vector <int> > stacgraphrev, set<int> sblks, set<int> MStops, map<int, double> *intersblkprob) {
	int numblks = stacinfo.stacgraph.size();
	for(set<int>::iterator miter = sblks.begin(); miter != sblks.end(); miter++) {
		set<int>::iterator tmpt = MStops.find(stacinfo.toplabs[*miter]);
		if(tmpt != MStops.end())	//this is not an intersblk
			continue;
		int memoflag[numblks];
		double memoprob[numblks];
		memset((char*)memoflag, '\0', sizeof(int)*numblks);
		memset((char*)memoprob, '\0', sizeof(double)*numblks);
		(*intersblkprob)[*miter] = dyndisjreachprob(&stacgraphrev, &stacinfo, *miter, &MStops, memoflag, memoprob);	//intersblkprobs + to [0][0]
	}
}

int getProjection (int theargsL, structureInfo stacinfo, vector < pair <int, pair <int, int> > > titersecond, map<int, vector<int> > *stacgraphproj, map<int, double> *intersblkprob) {
	int numblks = stacinfo.stacgraph.size();
	map <int, vector <int> > stacgraphrev;
	reverseGraph(&stacinfo.stacgraph, &stacgraphrev);
		
//1.get the minimal set of startblks that are < than others
//2.getprojn on all startblks, and identify the overall set of endblks
//3.get the minimal set of endblks that are > than others
//4.project the subgph frmo the single MS start blk (or dummy) to single MSE end blk (or dummy) -> Only 1 projection
//5.for each startblk not in MS, compute the prob of reaching it WITHOUT seeing any other startblk (if we do by dyn prog rev dfs, these are its ancestors)
//6.now do the matr multplcn; we can reuse the dyn prog based thingy or forwd bfs with priorq thingy; in either case, we need a vector to track intermed startblks
//7.when trying to mult parent's summed matr with curr matr in an intermed startblk, add the disjreachprob (computed earlier) to the [0][0] entry of the parents sumemd matr before mult
//now, finally, we'll get the result matr, and we can read [0][f] as result prob
//note that the cases where numstartblks is only 1, the query procsng can be as is; but, in the other case, need to alter and include dsijreachprobs etc
	int regularcaseflag;
	set <int> endblks;													//we get the endblks, and also the grph itself, as proj
	map <int, int> maxremlen;										//we send the maxremlen from each block ahead
	set <int> poststartblks;										//set of start blocks from postings -> used for var len query
	
	if(theargsL != -1) {		//so it is a fixlen query; obtain the list of startblks, maxremlen from that blk; then we'll use it to obtain endblks
		for(vector< pair <int, pair <int, int> > >::iterator piter = titersecond.begin(); piter != titersecond.end(); piter++) {
			int minlen = stacinfo.mmlens[piter->first].first;	//minpathlen for this blkno
			int calcremlen = theargsL - minlen + piter->second.second;	//this offsetno
			calcremlen = calcremlen < 0 ? 0 : calcremlen;		//cannot be negative
			map<int, int>::iterator tmpit = maxremlen.find(piter->first);
			if(tmpit != maxremlen.end()) {			//already seen this blkno before; so chk for max calcremlen
				if(maxremlen[piter->first] < calcremlen)
					maxremlen[piter->first] = calcremlen;
			}
			else	//seeing this blkno first time, assing this calcremlen
				maxremlen[piter->first] = calcremlen;
		}
		int numstartblks = maxremlen.size(); //chk each blk, remlen
		if(numstartblks == 1) {		//if only one startblk, simple - we only need to find the end nodes
			regularcaseflag = 1;
			map<int, int>::iterator fiter = maxremlen.begin();	//the only blk
			if(fiter->second == 0)	//maxremlen is 0; so, no need to look at any other block; this is the simplest case
				(*stacgraphproj)[fiter->first];
			else	//maxremlen > 0; so need to do bfs and handover to children
				getprojn(&stacinfo, &stacgraphrev, stacgraphproj, &endblks, fiter->second, fiter->first);		//get max consrtv subgph
		}
		else {	
			regularcaseflag = 2;
			set <int> thestartblks;
			for(map<int, int>::iterator fiter = maxremlen.begin(); fiter != maxremlen.end(); fiter++)
				thestartblks.insert(fiter->first);
			//1.get the minimal set of startblks that are < than others
			set <int> MStops;			//the sblks of the minimal MS set; this is wt we need to compute
			getMinSet(stacinfo, thestartblks, &MStops, 1);
			
			//2.getprojn on all startblks, and identify the overall set of endblks; note that we only need the set of endblks reached; not the subgph projection
			//if the getprojn fails to reach passonremlen 0, then the endblks set will be empty; though the subgrph wont be
			//this means the overall projn may be wrong (even empty) thus, in that case, we add the global endblks of the graph as endblks of that projn
			int globalendblksflag = 0;
			set <int> tendblks;			//the union of all sets of endlbks from getprojn; these are orig blknos
			for(map<int, int>::iterator miter = maxremlen.begin(); miter != maxremlen.end(); miter++) {
				map <int, vector <int> > junkgph;
				set <int> tmpendblks;
				getprojn(&stacinfo, &stacgraphrev, &junkgph, &tmpendblks, miter->second, miter->first);
				if(tmpendblks.size() == 0) //in case we get no endblks, it means we reached end of fst before passonrem became 0; so global endblks need to be included
					globalendblksflag = 1;
				for(set<int>::iterator titer = tmpendblks.begin(); titer != tmpendblks.end(); titer++)
					tendblks.insert(*titer);
			}
			if(globalendblksflag == 1) //obtain gloabal end blks and add to tnedblks
				for(map<int, vector<int> >::iterator geiter = stacinfo.stacgraph.begin(); geiter != stacinfo.stacgraph.end(); geiter++)
					if(geiter->second.size() == 0)		//no children, so a global endblk
						tendblks.insert(geiter->first);			//Now we have the whole set of endblks

			//3.get the minimal set of endblks that are > than others
			set <int> MSEtops;			//the eblks of the minimal MSE set; this is wt we need to compute
			getMinSet(stacinfo, tendblks, &MSEtops, -1); 		//dfssbi rev; limit rev since flag -1
					
			//4.project the subgph frmo the single MS start blk (or dummy) to single MSE end blk (or dummy) -> Only 1 projection on stacgph
			//since the MS are disjt and the MSE are disjt, we can get set of nodes that occur after MS and before MSE and intersect them and project stacgph
			int afterMS[numblks];
			int beforeMSE[numblks]; //not toplabs! these vectors cobntain origblknos!
			memset((char*)afterMS, '\0', sizeof(int)*numblks);
			memset((char*)beforeMSE, '\0', sizeof(int)*numblks);
			for(set<int>::iterator msiter = MStops.begin(); msiter != MStops.end(); msiter++)
				dfsgetreach(&stacinfo.stacgraph, stacinfo.toplabsrev[*msiter], afterMS);
			for(set<int>::iterator mseiter = MSEtops.begin(); mseiter != MSEtops.end(); mseiter++)
				dfsgetreach(&stacgraphrev, stacinfo.toplabsrev[*mseiter], beforeMSE);
			for(map<int, vector<int> >::iterator priter = stacinfo.stacgraph.begin(); priter != stacinfo.stacgraph.end(); priter++) {
				if((afterMS[priter->first] == 0) || (beforeMSE[priter->first] == 0))		//if this fromnode is not present in either of the presencevectors, ignore
					continue;
				for(vector<int>::iterator viter = priter->second.begin(); viter != priter->second.end(); viter++) {
					if((afterMS[*viter] == 0) || (beforeMSE[*viter] == 0))
						continue;	//both from and to nodes are in subgph
					(*stacgraphproj)[priter->first].push_back(*viter);	//add edge
					(*stacgraphproj)[*viter];	//simply mention this to include in gph
				}
			}
			//5.for each startblk not in MS, compute the prob of reaching it without seeing any other startblk 			
			getInterProb(stacinfo, stacgraphrev, thestartblks, MStops, intersblkprob);
		} //end if on numstartblks
	} //end if Lfixlen
	
	else {	//var len; we need to project till end of blkgph
		for(vector< pair <int, pair <int, int> > >::iterator piter = titersecond.begin(); piter != titersecond.end(); piter++)
			poststartblks.insert(piter->first);
		int numstartblks = poststartblks.size(); 		//chk each blk, remlen
		if(numstartblks == 1) {
			regularcaseflag = 1;		//for varlen, we're anyways projecting till end
			int allcdone[numblks];			//Memo to ntoe wheteher all children of a node have been visited! in that case, we neednt repeat this
			memset((char*)allcdone, '\0', sizeof(int)*numblks);	//without this memo, the runtime could be O(V!) and with it, it is O(V+E)
			dfstillend(&stacinfo.stacgraph, stacgraphproj, allcdone, *(poststartblks.begin()));
		}
		else {
			regularcaseflag = 2;
			//1.get the minimal set of startblks that are < than others
			set <int> MStops;			//the sblks of the minimal MS set; this is wt we need to compute
			getMinSet(stacinfo, poststartblks, &MStops, 1);

			//Now we need to project it till end by dfs
			int afterMS[numblks];
			memset((char*)afterMS, '\0', sizeof(int)*numblks);
			for(set<int>::iterator msiter = MStops.begin(); msiter != MStops.end(); msiter++)
				dfsgetreach(&stacinfo.stacgraph, stacinfo.toplabsrev[*msiter], afterMS);
			for(map<int, vector<int> >::iterator priter = stacinfo.stacgraph.begin(); priter != stacinfo.stacgraph.end(); priter++) {
				if(afterMS[priter->first] == 0) //if this fromnode is not present in either of the presencevectors, ignore
					continue;
				for(vector<int>::iterator viter = priter->second.begin(); viter != priter->second.end(); viter++) {
					if(afterMS[*viter] == 0)
						continue;		//both from and to nodes are in subgph
					(*stacgraphproj)[priter->first].push_back(*viter);		//add edge
					(*stacgraphproj)[*viter];	//simply mention this to include in gph
				}				
			}	
			//5.for each startblk not in MS, compute the prob of reaching it without seeing any other startblk 			
			getInterProb(stacinfo, stacgraphrev, poststartblks, MStops, intersblkprob);
		}	//end numsblks
	}	//end if on fixlen vs varlen
	
	return regularcaseflag;
}
