/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "frontends/staccatoSQLUtils.h"

//incrementally add the inp matr to the dest matr; matr size is NUMST + NUMST
void incrsummatrs (double* dest, double *inp, int NUMST) {
	for(int i=0; i < (NUMST * NUMST); i++)
		dest[i] += inp[i];			//doesnt matter if row/col major
}

void multmatrs(double* dest, double* m1, double* m2, int NUMST) {
	for(int i=0; i<NUMST; i++)
		for(int j=0; j<NUMST; j++) {
			dest[i*NUMST + j] = 0.0;
			for(int k=0; k<NUMST; k++)
				dest[i*NUMST + j] += m1[i*NUMST + k] * m2[k*NUMST + j];		//row major, fully thro
		}
}

void printmatr(double* m, int NUMST) {
	for(int i=0; i<NUMST; i++) {
		cout << "i:" << i << ":\t";
		for(int j=0; j<NUMST; j++)
			cout << m[i*NUMST + j] << "\t";
		cout << endl;
	}	
}

void scalarmultmatr (double* dest, int NUMST, double scalprob) {
	for(int i=0; i < (NUMST * NUMST); i++)
		dest[i] *= scalprob;			//doesnt matter if row/col major
}

//this function uses a pre-allocated memory for the matrs, and uses scratches - to store sum of children
//NUMST is the length/breadth of the probmatr; and there are numblks such matrs, and numbkls such ints in the memo
//curnode is the blknum we are at;
void matrdfs (double* probmatrs, double* scratches, int* memo, int NUMST, map <int, vector <int> > *stacgraph, int curnode) {
	if (memo[curnode] == 1)
		return;
	if((*stacgraph)[curnode].size() == 0) {
		memo[curnode] = 1;
		return;
	}
	for(vector <int>::iterator iter = (*stacgraph)[curnode].begin(); iter != (*stacgraph)[curnode].end(); iter++)
		matrdfs (probmatrs, scratches, memo, NUMST, stacgraph, *iter);
	int matrsize = NUMST * NUMST;	//size of matr in num doubles
	memset(scratches, 0, 2 * matrsize * sizeof(double));	//size of each prob matr is NUMST * NUMST * sizeofdouble in bytes
	for(vector <int>::iterator iter = (*stacgraph)[curnode].begin(); iter != (*stacgraph)[curnode].end(); iter++)
		incrsummatrs(scratches, probmatrs + ((*iter) * matrsize), NUMST);	//add this child's probmatr to the scratch0
	multmatrs(scratches + matrsize, probmatrs + (curnode * matrsize), scratches, NUMST);	//mult is not commutative
	memcpy(probmatrs + (curnode * matrsize), scratches + matrsize, matrsize * sizeof(double));	//size of amtr is matrsize * sizeofdouble in bytes
	memo[curnode] = 1;
	return;
}

void graphread(structureInfo *stacinfo, char* buffer) {
	int sizeofint = sizeof(int);
	int sizeofdouble = sizeof(double);
	int numblks = 0, numarcs = 0;
	memcpy((char*)&numblks, buffer, sizeofint);
	memcpy((char*)&numarcs, buffer+sizeofint, sizeofint);
	int addendum = 2*sizeofint;
	for(int i=0; i<numblks; i++) {
		(stacinfo->toplabs).push_back(0);
		(stacinfo->toplabsrev).push_back(0);
		(stacinfo->blkprobs).push_back(0);
		(stacinfo->blkreachprobs).push_back(0);
	}
	for(int i=0; i<numblks; i++) {
		(stacinfo->stacgraph)[i];
		double bwt =0.0, rwt = 0.0;
		int from = 0, topl = 0, minl = 0, maxl = 0, numtos = 0;	
		memcpy((char*)&from, buffer+addendum, sizeofint);
		memcpy((char*)&topl, buffer+addendum+sizeofint, sizeofint);
		memcpy((char*)&bwt, buffer+addendum+2*sizeofint, sizeofdouble);
		memcpy((char*)&rwt, buffer+addendum+2*sizeofint+sizeofdouble, sizeofdouble);
		(stacinfo->toplabs)[from] = topl;
		(stacinfo->toplabsrev)[topl] = from;
		(stacinfo->blkprobs)[from] = exp(-1.0 * bwt);
		(stacinfo->blkreachprobs)[from] = exp(-1.0 * rwt);
		addendum += 2 * (sizeofint + sizeofdouble);
		memcpy((char*)&minl, buffer+addendum, sizeofint);
		memcpy((char*)&maxl, buffer+addendum+sizeofint, sizeofint);
		memcpy((char*)&numtos, buffer+addendum+2*sizeofint, sizeofint);
		addendum += 3*sizeofint;
		if(minl < 1)
			minl = 1;
		(stacinfo->mmlens)[from] = make_pair(minl, maxl);
		for(int x=0; x<numtos; x++) {
			int to = 0;
			memcpy((char*)&to, buffer+addendum, sizeofint);
			addendum += sizeofint;
			(stacinfo->stacgraph)[from].push_back(to);
		}
	}
}

void graphwrite(structureInfo *stacinfo, ostream &fout) {
	//<numnodes> <numarcs> <fromblk> <fromblk topsort label> <blkreachprob in -log> <minpathlen> <maxpathlen> <numtoblks> <to1> <to2> <to3> ... <fromblk>...
	int sizeofint = sizeof(int);
	int sizeofdouble = sizeof(double);
	int numblks = (stacinfo->stacgraph).size();
	fout.write((char*)&numblks, sizeofint);
	int numarcs = 0;
	for(map <int, vector <int> >::iterator iter = (stacinfo->stacgraph).begin(); iter != (stacinfo->stacgraph).end(); iter++) {
		numarcs += iter->second.size();
	}
	fout.write((char*)&numarcs, sizeofint);
	for(map <int, vector <int> >::iterator iter = (stacinfo->stacgraph).begin(); iter != (stacinfo->stacgraph).end(); iter++) {
		int from = iter->first;
		double bwt = -1.0 * log(stacinfo->blkprobs[from]);		//we store the wt, not the prob
		double rwt = -1.0 * log(stacinfo->blkreachprobs[from]);		//we store the wt, not the prob
		fout.write((char*)&from, sizeofint);
		fout.write((char*)&(stacinfo->toplabs[from]), sizeofint);
		fout.write((char*)&(bwt), sizeofdouble);
		fout.write((char*)&(rwt), sizeofdouble);
		fout.write((char*)&((stacinfo->mmlens)[from].first), sizeofint);
		fout.write((char*)&((stacinfo->mmlens)[from].second), sizeofint);
		int numtos = iter->second.size();
		fout.write((char*)&numtos, sizeofint);
		for(vector <int>::iterator siter = iter->second.begin(); siter != iter->second.end(); siter++) {
			int to = *siter;
			fout.write((char*)&to, sizeofint);
		}
	}
}

void reverseGraph(map <int, vector <int> > *gph, map <int, vector <int> > *revgph) {
	for(map <int, vector <int> >::iterator iter = gph->begin(); iter != gph->end(); iter++) {
		(*revgph)[iter->first];
		for(vector <int>::iterator siter = iter->second.begin(); siter != iter->second.end(); siter++)
			 (*revgph)[*siter].push_back(iter->first);
	}
}

void sortArray(int* array, int arlen) {
	for(int x=0; x<arlen-1; x++)
		for(int y=x+1; y<arlen; y++)
			if(array[x] > array[y]) {
				int tmps = array[x];
				array[x] = array[y];
				array[y] = tmps;
			}
}
