/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libraries/NodeEdge.h"

Node::Node() {
	this->index = -2;
	this->in.clear();
	this->out.clear();
}

Node::Node(int index, vector<EdgeWeight *> in, vector<EdgeWeight *> out) {
	this->index = index;
	this->in.assign(in.begin(), in.end());
	this->out.assign(out.begin(), out.end());
}

Node::~Node() {
	this->in.clear();
	this->out.clear();
}

Node& Node::operator=(Node &rhs) {
	this->index = rhs.index;
	this->in.assign(rhs.in.begin(), rhs.in.end());
	this->out.assign(rhs.out.begin(), rhs.out.end());
    return (*this);
}

EdgeWeight::EdgeWeight() {
	this->from = -2;
	this->to = -2;
	this->weight = 0;
}

EdgeWeight::EdgeWeight(int from, int to, const double &value, string label, int of, int ot) {
	this->from = from;
	this->to = to;
	this->weight = value;
   this->label = label;
	this->origfrom = of;
	this->origto = ot;
}

bool EdgeWeight::operator==(const EdgeWeight &rhs) {
    return this->from == rhs.from
        && this->to == rhs.to
        && this->label == rhs.label;
}

bool EdgeWeight::operator<(const EdgeWeight &rhs) {
    return (this->from < rhs.from)
        || (this->from == rhs.from && this->to < rhs.to)
        || (this->from == rhs.from && this->to == rhs.to
            && this->label < rhs.label);
}
