/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <sstream>
#include <vector>
#include <cstdlib>
#include "libpq-fe.h"
using namespace std;

int main(int argc, char** argv) {
	if(argc < 7) {
		cout << "Usage: loadindex <dbname> <port number> <index table> <docname> <mbinname> <.index file>\n";
		return 0;
	}
	PGconn* conn;
	string dbstr = "dbname=" +  string(argv[1]) + " port=" + string(argv[2]);
	conn = PQconnectdb(dbstr.c_str());
	if(PQstatus(conn) != CONNECTION_OK) {
		cout << "Error: Could not connect with " << dbstr << "! Message: " << PQerrorMessage(conn);
		PQfinish(conn);
		exit(1);
	}
	ifstream fin(argv[6], ios::in);
	if(!fin.is_open()) {
		cout << "Error: Cannot open index file " << argv[6] << endl;
		PQfinish(conn);
		return 1;
	}
	string tablename = string(argv[3]);
	string docname = string(argv[4]);
	string fstname = string(argv[5]);
	string pathkey = docname + "-" + fstname;
	string comnd = "insert into " + tablename + " values ";
	vector <string> insvals;		//to be used to buffer writes to the database -> we send about 100 values at one write
	int WRITESIZE = 100;
	while(!fin.eof()) {
		string indexline;
		getline(fin, indexline);
		if(indexline.compare("") == 0)	//empty string
			continue;
		char* tok1 = strtok((char*)indexline.c_str(), "\t");	//removing the \t:\t between the term and postings
		vector <string> toks1;
		while(tok1 != NULL) {
			toks1.push_back(tok1);
			tok1 = strtok(NULL, "\t");
		}
		string term = toks1[0];
		char* pos = strtok((char*)toks1[2].c_str(), " ");		//isolate the postings
		if(pos == NULL)
			continue;
		int numpos = atoi(pos);
		vector <int> posints;
		int numints = 0;
		pos = strtok(NULL, " ");
		while(pos != NULL) {
			posints.push_back(atoi(pos));
			pos = strtok(NULL, " ");
			numints++;
		}
		if(numints != 3*numpos) {
			cout << "Error: For term " << term << " number of postings is " << numpos << " but entry in file contains only " << numints << " integers (!=" << 3*numpos << ")!" << endl;
			return 1;
		}
		string thisterm = "";
		for(int i=0; i<numpos; i++) {
			//create the query for this term this pos -> assuming that the terms do NOT have \, (, ), ', "!! -> dict.txt doesnt
			stringstream blk, pth, off;
			blk << posints[3*i];
			pth << posints[3*i + 1];
			off << posints[3*i + 2];
			string thispos = "('" + term + "', '" + pathkey + "', " + blk.str() + ", " + pth.str() + ", " + off.str() + ")";
			insvals.push_back(thispos);
			if(insvals.size() >= WRITESIZE) {
				string flushcmd = "insert into " + tablename + " values " + insvals[0];
				for(int z=1; z<insvals.size(); z++)
					flushcmd += ", " + insvals[z];
				PGresult *res;
				res = PQexec(conn, flushcmd.c_str());
				PQclear(res);
				insvals.clear();
			}
		}
	}
	if(insvals.size() != 0) { //flush to databse the remng entries
		string flushcmd = "insert into " + tablename + " values " + insvals[0];
		for(int z=1; z<insvals.size(); z++) {
			flushcmd += ", " + insvals[z];
		}
		PGresult *res;
		res = PQexec(conn, flushcmd.c_str());
		PQclear(res);
		insvals.clear();
	}
	PQfinish(conn);

return 0;
}
